/*  Basic m68hc05 simulator
    Copyright 2004 Free Software Foundation, Inc.
    Copyright 2004 Martin Kgler
    Contributed by Martin Kgler (mkoegler@auto.tuwien.ac.at)
    (based on m68hc11-run)
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
    
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street - Fifth Floor, Boston, MA 02110-1301, USA.
    
    */

#include "sim-main.h"
#include "sim-assert.h"
#include "sim-module.h"
#include "sim-options.h"
#include "sim-io.h"

enum
{
  OPTION_CPU_RESET = OPTION_START,
  OPTION_EMUL_OS,
  OPTION_CPU_MODE
};

static DECLARE_OPTION_HANDLER (cpu_option_handler);

static const OPTION cpu_options[] = {
  {{"cpu-reset", no_argument, NULL, OPTION_CPU_RESET},
   '\0', NULL, "Reset the CPU",
   cpu_option_handler},

  {{"no-emulos", no_argument, NULL, OPTION_EMUL_OS},
   '\0', NULL, "Emulate no OS system calls",
   cpu_option_handler},

  {{NULL, no_argument, NULL, 0}, '\0', NULL, NULL, NULL}
};

static SIM_RC
cpu_option_handler (SIM_DESC sd, sim_cpu * cpu,
		    int opt, char *arg, int is_command)
{
  int val;

  cpu = STATE_CPU (sd, 0);
  switch (opt)
    {
    case OPTION_CPU_RESET:
      sim_board_reset (sd);
      break;

    case OPTION_EMUL_OS:
      cpu->cpu_emul_syscall = 0;
      break;

    case OPTION_CPU_MODE:
      break;
    }

  return SIM_RC_OK;
}

void
cpu_call (sim_cpu * cpu, uint16 addr)
{

  cpu_set_pc (cpu, addr);
}

void
cpu_return (sim_cpu * cpu)
{
}

void
cpu_single_step (sim_cpu * cpu)
{
  cpu->cpu_current_cycle = 0;
  cpu->cpu_insn_pc = cpu_get_pc (cpu);

  /* Handle the pending interrupts.  If an interrupt is handled,
     treat this as an single step.  */
  if (interrupts_process (&cpu->cpu_interrupts))
    {
      cpu->cpu_absolute_cycle += cpu->cpu_current_cycle;
      return;
    }

  /*  printf("PC = 0x%04x\n", cpu_get_pc (cpu)); */
  cpu->cpu_interpretor (cpu);
  cpu->cpu_absolute_cycle += cpu->cpu_current_cycle;
}

int
cpu_initialize (SIM_DESC sd, sim_cpu * cpu)
{
  sim_add_option_table (sd, 0, cpu_options);

  memset (&cpu->cpu_regs, 0, sizeof (cpu->cpu_regs));

  cpu->cpu_absolute_cycle = 0;
  cpu->cpu_current_cycle = 0;
  cpu->cpu_emul_syscall = 1;
  cpu->cpu_running = 1;
  cpu->cpu_stop_on_interrupt = 0;
  cpu->cpu_frequency = 8 * 1000 * 1000;
  cpu->cpu_use_elf_start = 0;
  cpu->cpu_elf_start = 0;

  interrupts_initialize (sd, cpu);

  cpu->cpu_is_initialized = 1;
  return 0;
}

/* Reinitialize the processor after a reset.  */
int
cpu_reset (sim_cpu * cpu)
{
  /* Setup the processor registers.  */
  memset (&cpu->cpu_regs, 0, sizeof (cpu->cpu_regs));
  cpu->cpu_absolute_cycle = 0;
  cpu->cpu_current_cycle = 0;
  cpu->cpu_is_initialized = 0;

  /* Reset interrupts.  */
  interrupts_reset (&cpu->cpu_interrupts);

  return 0;
}

/* Reinitialize the processor after a reset.  */
int
cpu_restart (sim_cpu * cpu)
{
  uint16 addr;

  /* Get CPU starting address depending on the CPU mode.  */
  if (cpu->cpu_use_elf_start == 0)
    {
      addr = memory_read16 (cpu, 0xFFFE);
    }
  else
    {
      addr = cpu->cpu_elf_start;
    }

  /* Setup the processor registers.  */
  cpu->cpu_insn_pc = addr;
  cpu->cpu_regs.pc = addr;
  cpu->cpu_regs.ccr = 0;
  cpu->cpu_regs.sp = 0xff;
  cpu->cpu_regs.a = 0;
  cpu->cpu_regs.x = 0;
  cpu->cpu_absolute_cycle = 0;
  cpu->cpu_is_initialized = 1;
  cpu->cpu_current_cycle = 0;

  cpu_call (cpu, addr);

  return 0;
}

void
cpu_info (SIM_DESC sd, sim_cpu * cpu)
{
  sim_io_printf (sd, "CPU info:\n");
  sim_io_printf (sd, "  Absolute cycle: %s\n",
		 cycle_to_string (cpu, cpu->cpu_absolute_cycle,
				  PRINT_TIME | PRINT_CYCLE));

  sim_io_printf (sd, "  Syscall emulation: %s\n",
		 cpu->cpu_emul_syscall ? "yes, via 0xcd <n>" : "no");
  sim_io_printf (sd, "  Memory errors detection: %s\n",
		 cpu->cpu_check_memory ? "yes" : "no");
  sim_io_printf (sd, "  Stop on interrupt: %s\n",
		 cpu->cpu_stop_on_interrupt ? "yes" : "no");
}

void
cpu_memory_exception (sim_cpu * cpu, SIM_SIGNAL excep,
		      uint16 addr, const char *message)
{
  if (cpu->cpu_running == 0)
    return;

  cpu_set_pc (cpu, cpu->cpu_insn_pc);
  sim_engine_halt (CPU_STATE (cpu), cpu, NULL,
		   cpu_get_pc (cpu), sim_stopped, excep);

}

/* Handle special instructions.  */
void
cpu_special (sim_cpu * cpu, uint8 special)
{
  char c;
  int i;
  switch (special)
    {
    case 0x83:
      if (cpu->cpu_emul_syscall)
	{
	  switch (cpu_get_x (cpu))
	    {
	    case 0:
	      sim_engine_halt (CPU_STATE (cpu), cpu,
			       NULL, NULL_CIA, sim_exited, cpu_get_a (cpu));
	      break;
	    case 1:
	      c = cpu_get_a (cpu);
	      sim_io_write_stdout (CPU_STATE (cpu), &c, 1);
	      break;
	    case 2:
	      i = sim_io_read_stdin (CPU_STATE (cpu), &c, 1);
	      if (i == 1)
		{
		  cpu_set_ccr_C (cpu, 1);
		  cpu_set_a (cpu, c);
		}
	      else
		cpu_set_ccr_C (cpu, 0);
	      break;

	    default:
	      sim_engine_abort (CPU_STATE (cpu), cpu, 0,
				"Invalid syscall 0x%0x", cpu_get_x (cpu));
	    }
	  cpu_add_cycles (cpu, 100);
	  break;
	}
    case 0xA7:
      {
	SIM_DESC sd;

	sd = CPU_STATE (cpu);

	/* Breakpoint instruction if we are under gdb.  */
	if (STATE_OPEN_KIND (sd) == SIM_OPEN_DEBUG)
	  {
	    cpu->cpu_regs.pc--;
	    sim_engine_halt (CPU_STATE (cpu), cpu,
			     0, cpu_get_pc (cpu), sim_stopped, SIM_SIGTRAP);
	  }
      }
    default:
      cpu_set_pc (cpu, cpu->cpu_insn_pc);
      sim_engine_halt (CPU_STATE (cpu), cpu, NULL,
		       cpu_get_pc (cpu), sim_stopped, SIM_SIGILL);
      return;
    }
}
